'''
[CycleGAN]
Jun-Yan Zhu, et al. Unpaired Image-to-Image Translation using Cycle-Consistent Adversarial Networks. ICCV 2017.
Code reference: https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix

[运行]
python Practice_20_CycleGAN.py

[环境要求]
- Pytorch
- torchvision
- opencv
- matplotlib
- pillow
- tqdm

[数据]
在本文件夹的data/unpaired目录下下载以下数据用作训练
wget https://people.eecs.berkeley.edu/~taesung_park/CycleGAN/datasets/horse2zebra.zip --no-check-certificate
unzip -q horse2zebra.zip
rm -f horse2zebra.zip
然后将文件结构整理如下
data/unpaired/horse2zebra
├── testA
│   ├── n02381460_20.jpg
│   ├── ...
│   └── n02381460_9260.jpg
├── testB
│   ├── n02391049_80.jpg
│   ├── ...
│   └── n02391049_10980.jpg
├── dataA/train
│   ├── n02381460_2.jpg
│   ├── ...
│   └── n02381460_9263.jpg
└── dataB/train
    ├── n02391049_2.jpg
    ├── ...
    └── n02391049_11195.jpg
'''

import os
from PIL import Image
import matplotlib.pyplot as plt

import torchvision.transforms as transforms
import torchvision.models as models

import numpy as np
from datetime import datetime
import time
import logging
import random
import cv2
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.autograd as autograd
from torch.autograd import Variable
from torch.nn import Linear, Conv2d, BatchNorm2d, LayerNorm, LeakyReLU, ConvTranspose2d, ReLU, Tanh, InstanceNorm2d
from torch.nn import ReflectionPad2d, ReplicationPad2d

from torch.utils.data import DataLoader
import torchvision.datasets as dset
import torchvision
import itertools
from tqdm import tqdm


'''辅助函数''' 
# 读图像
def load_image(filename):
    transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(mean=[0.5, 0.5, 0.5],std=[0.5,0.5,0.5]),
    ])
    
    img = Image.open(filename)
    img = transform(img)
    return img.unsqueeze(dim=0)   

# 保存图像
def save_image(img, filename):
    tmp = ((img.detach().numpy().transpose(1, 2, 0) + 1.0) * 127.5).astype(np.uint8)
    cv2.imwrite(filename, cv2.cvtColor(tmp, cv2.COLOR_RGB2BGR))

# 网络参数初始化
def weights_init(m):
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
        if not(m.bias is None):
            m.bias.data.fill_(0)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)
        
        
# 定义计算对抗损失的GANLoss类
class GANLoss(nn.Module):
    def __init__(self, LSGAN=True):
        super(GANLoss, self).__init__()
        # 定义两个目标张量，分别表示True和False的1.0和0.0
        self.register_buffer('real_label', torch.tensor(1.0))
        self.register_buffer('fake_label', torch.tensor(0.0))
        if LSGAN: 
            # LSGAN loss
            self.loss = nn.MSELoss()
        else:
            # 原版loss
            self.loss = nn.BCEWithLogitsLoss()

    def forward(self, prediction, target_is_real):
        if target_is_real:
            target_tensor = self.real_label
        else:
            target_tensor = self.fake_label
        # 将目标张量扩展到与输出的特征的维度一致
        target_tensor = target_tensor.expand_as(prediction)
        return self.loss(prediction, target_tensor)

    
'''网络定义'''
# 定义残差模块
class ResnetBlock(nn.Module):
    def __init__(self, dim):
        super(ResnetBlock, self).__init__()

        conv_block = []
        conv_block += [nn.ReflectionPad2d(1)]
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=0)]
        conv_block += [nn.InstanceNorm2d(dim), nn.ReLU(True)]
        conv_block += [nn.ReflectionPad2d(1)]
        conv_block += [nn.Conv2d(dim, dim, kernel_size=3, padding=0)]
        conv_block += [nn.InstanceNorm2d(dim)]

        self.conv_block = nn.Sequential(*conv_block)
        self.relu = nn.ReLU(True)

    def forward(self, x):
        out = x + self.conv_block(x)  
        return self.relu(out)

class ResnetGenerator(nn.Module):
    def __init__(self, ngf=64, n_blocks=6):
        super(ResnetGenerator, self).__init__()
        # 编码器，包含1个7*7的卷积层和2个3*3的下采卷积层
        model = [nn.ReflectionPad2d(3),
                 nn.Conv2d(3, ngf, kernel_size=7, padding=0),
                 nn.InstanceNorm2d(ngf),
                 nn.ReLU(True)]

        n_downsampling = 2
        for i in range(n_downsampling): 
            mult = 2 ** i
            model += [nn.Conv2d(ngf * mult, ngf * mult * 2, kernel_size=3, stride=2, padding=1),
                      nn.InstanceNorm2d(ngf * mult * 2),
                      nn.ReLU(True)]

        # 中间的n_blocks层残差模块
        mult = 2 ** n_downsampling
        for i in range(n_blocks):       
            model += [ResnetBlock(ngf * mult)]

        # 解码器，包含2个3*3的上采反卷积层和1个7*7的卷积层和
        for i in range(n_downsampling): 
            mult = 2 ** (n_downsampling - i)
            model += [nn.ConvTranspose2d(ngf * mult, int(ngf * mult / 2),
                                         kernel_size=3, stride=2,
                                         padding=1, output_padding=1),
                      nn.InstanceNorm2d(int(ngf * mult / 2)),
                      nn.ReLU(True)]
        model += [nn.ReflectionPad2d(3)]
        model += [nn.Conv2d(ngf, 3, kernel_size=7, padding=0)]
        model += [nn.Tanh()]

        self.model = nn.Sequential(*model)

    def forward(self, input):
        return self.model(input)
    
print(ResnetGenerator(ngf=64, n_blocks=9))

class NLayerDiscriminator(nn.Module):
    def __init__(self, ndf=64, n_layers=3):
        super(NLayerDiscriminator, self).__init__()
        sequence = [
            nn.Conv2d(3, ndf, kernel_size=4, stride=2, padding=1),
            nn.LeakyReLU(0.2, True)
        ]
        nf_mult = 1
        nf_mult_prev = 1
        # 随着层数的加深逐渐增加卷积核的数量
        for n in range(1, n_layers):  
            nf_mult_prev = nf_mult
            nf_mult = min(2 ** n, 8)
            sequence += [
                nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=4, stride=2, padding=1),
                nn.InstanceNorm2d(ndf * nf_mult),
                nn.LeakyReLU(0.2, True)
            ]

        nf_mult_prev = nf_mult
        nf_mult = min(2 ** n_layers, 8)
        sequence += [
            nn.Conv2d(ndf * nf_mult_prev, ndf * nf_mult, kernel_size=4, stride=1, padding=1),
            nn.InstanceNorm2d(ndf * nf_mult),
            nn.LeakyReLU(0.2, True)
        ]

        # 输出层的通道数降为1，空间坐标上的每个元素指示真假
        sequence += [nn.Conv2d(ndf * nf_mult, 1, kernel_size=4, stride=1, padding=1)]  
        self.model = nn.Sequential(*sequence)

    def forward(self, input):
        return self.model(input)

# 定义CycleGAN网络模型及训练框架
class CycleGANModel(nn.Module):
    def __init__(self):
        super(CycleGANModel, self).__init__()
        # 定义G网络包含9层Resblock的编码器-Resblocks-解码器结构
        # 定义D网络，包含3+2层卷积层
        self.netG_A = ResnetGenerator(ngf=64, n_blocks=9)
        self.netG_B = ResnetGenerator(ngf=64, n_blocks=9)
        self.netD_A = NLayerDiscriminator(n_layers=3, ndf=64)
        self.netD_B = NLayerDiscriminator(n_layers=3, ndf=64)

        # 定义对抗损失和L1重建损失
        self.criterionGAN = GANLoss()
        self.criterionCycle = torch.nn.L1Loss()
        # 定义G和D的优化器，itertools.chain使两个网络参数一起更新
        self.optimizer_G = torch.optim.Adam(itertools.chain(self.netG_A.parameters(), self.netG_B.parameters()), lr=0.0002, betas=(0.5, 0.999))
        self.optimizer_D = torch.optim.Adam(itertools.chain(self.netD_A.parameters(), self.netD_B.parameters()), lr=0.0002, betas=(0.5, 0.999))

    # G_A和G_B将A映射为B再重建A，并将B映射为A再重建B
    def forward(self, real_A, real_B):
        fake_B = self.netG_A(real_A) # G_A(A)
        rec_A = self.netG_B(fake_B)  # G_B(G_A(A))
        fake_A = self.netG_B(real_B) # G_B(B)
        rec_B = self.netG_A(fake_A)  # G_A(G_B(B))
        return [fake_B, rec_A, fake_A, rec_B]

    # 计算D网络的梯度
    def backward_D(self, netD, real, fake):
        # 对于真实图片，D网络需要判断其为True
        pred_real = netD(real)
        loss_D_real = self.criterionGAN(pred_real, True)  # 公式(6.5) (6.6)第一项
        # 对于生成图片，D网络需要判断其为False
        pred_fake = netD(fake.detach())
        loss_D_fake = self.criterionGAN(pred_fake, False) # 公式(6.5) (6.6)第二项
        loss_D = (loss_D_real + loss_D_fake) * 0.5
        loss_D.backward()                                  # 计算梯度并反向传播
        return loss_D.data.mean().item()

    # 计算G网络的梯度
    def backward_G(self, real_A, real_B, fake_B, rec_A, fake_A, rec_B):
        # 公式(6.5)第二项
        loss_G_A = self.criterionGAN(self.netD_A(fake_B), True)
        # 公式(6.6)第二项
        loss_G_B = self.criterionGAN(self.netD_B(fake_A), True)
        # 循环一致性损失，对应公式(6.7)第一项 || G_B(G_A(A)) - A||
        loss_cycle_A = self.criterionCycle(rec_A, real_A) * 10.0
        # 循环一致性损失，对应公式(6.7)第二项 || G_A(G_B(B)) - B||
        loss_cycle_B = self.criterionCycle(rec_B, real_B) * 10.0
        loss_G = loss_G_A + loss_G_B + loss_cycle_A + loss_cycle_B
        loss_G.backward()  # 计算梯度并反向传播
        return loss_G_A.data.mean().item(), loss_G_B.data.mean().item(), loss_cycle_A.data.mean().item(), loss_cycle_B.data.mean().item()


    # 训练流程，real_A和real_B为输入图像
    def optimize_parameters(self, real_A, real_B):
        [fake_B, rec_A, fake_A, rec_B] = self.forward(real_A, real_B)      
        # 更新G网络
        self.optimizer_G.zero_grad()  # 将G网络参数梯度初始化为0
        loss_G_A, loss_G_B, loss_cycle_A, loss_cycle_B = self.backward_G(real_A, real_B, fake_B, rec_A, fake_A, rec_B)        
        self.optimizer_G.step()       # 更新G网络参数
        # 更新D网络
        self.optimizer_D.zero_grad()  # 将D网络参数梯度初始化为0
        loss_D_A = self.backward_D(self.netD_A, real_B, fake_B.detach()) 
        loss_D_B = self.backward_D(self.netD_B, real_A, fake_A.detach()) 
        self.optimizer_D.step()       # 更新D网络参数
        return fake_B, rec_A, fake_A, rec_B, loss_G_A, loss_G_B, loss_cycle_A, loss_cycle_B, loss_D_A, loss_D_B
    
    
'''数据准备''' 
datasetA = dset.ImageFolder(root='./data/unpaired/horse2zebra/dataA/',
                           transform=transforms.Compose([
                               transforms.ToTensor(),
                               transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                           ]))

dataloaderA = DataLoader(datasetA, batch_size=1, shuffle=True, num_workers=2, drop_last=True)

datasetB = dset.ImageFolder(root='./data/unpaired/horse2zebra/dataB/',
                           transform=transforms.Compose([
                               transforms.ToTensor(),
                               transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
                           ]))

dataloaderB = DataLoader(datasetB, batch_size=1, shuffle=True, num_workers=2, drop_last=True)

print('DatasetA size:', len(dataloaderA.dataset.imgs), '; DatasetB size:', len(dataloaderB.dataset.imgs))


'''构建和初始化模型'''
model = CycleGANModel()
if torch.cuda.is_available():
    model.cuda()
model.netG_A.apply(weights_init)
model.netD_A.apply(weights_init)
model.netG_B.apply(weights_init)
model.netD_B.apply(weights_init)
print('Model created.') 


'''网络训练'''
pbar = tqdm(range(100000), initial=0, smoothing=0.01, dynamic_ncols=True)

iteratorA = iter(dataloaderA)
iteratorB = iter(dataloaderB)
for idx in pbar:
    try:
        data, _ = next(iteratorA)
    except StopIteration:
        iteratorA = iter(dataloaderA)
        data, _ = next(iteratorA) 
    real_A = data.cuda()
    try:
        data, _ = next(iteratorB)
    except StopIteration:
        iteratorB = iter(dataloaderB)
        data, _ = next(iteratorB) 
    real_B = data.cuda()
    
    [fake_B, rec_A, fake_A, rec_B, loss_G_A, loss_G_B, loss_cycle_A, 
     loss_cycle_B, loss_D_A, loss_D_B] = model.optimize_parameters(real_A, real_B)
        
    pbar.set_description(
        (
            f"dA: {loss_D_A:.3f}; dB: {loss_D_B:.3f}; gA: {loss_G_A:.3f}; gA: {loss_G_B:.3f}; "
            f"ca: {loss_cycle_A:.3f}; cb: {loss_cycle_B:.3f};"
        )
    )    

# 保存模型    
torch.save(model.state_dict(),  f"cyclegan-%05d.pt"%(idx))


'''测试模型'''
model.eval()
As = ['n02381460_490.jpg', 'n02381460_4630.jpg', 'n02381460_4160.jpg', 'n02381460_1870.jpg']
Bs = ['n02391049_4110.jpg', 'n02391049_1950.jpg', 'n02391049_3270.jpg', 'n02391049_3290.jpg']
model.eval()
for i in range(4):
    I_A = load_image('./data/unpaired/horse2zebra/testA/' + As[i])
    I_B = load_image('./data/unpaired/horse2zebra/testB/' + Bs[i])
    with torch.no_grad():
        fake_B, rec_A, fake_A, rec_B = model(I_A.cuda(), I_B.cuda())
    save_image(fake_B[0].detach().cpu(), '%s_A2B.jpg'%(As[i]))
    save_image(rec_A[0].detach().cpu(), '%s_A2B2A.jpg'%(As[i]))
    save_image(fake_A[0].detach().cpu(), '%s_B2A.jpg'%(Bs[i]))
    save_image(rec_B[0].detach().cpu(), '%s_B2A2B.jpg'%(Bs[i]))